package ginext

import (
	"github.com/gin-gonic/gin"
	"gogs.mikescher.com/BlackForestBytes/goext/langext"
	"net/http"
	"reflect"
	"runtime"
	"strings"
)

var anyMethods = []string{
	http.MethodGet, http.MethodPost, http.MethodPut, http.MethodPatch,
	http.MethodHead, http.MethodOptions, http.MethodDelete, http.MethodConnect,
	http.MethodTrace,
}

type GinRoutesWrapper struct {
	wrapper        *GinWrapper
	routes         gin.IRouter
	defaultHandler []gin.HandlerFunc
}

type GinRouteBuilder struct {
	routes *GinRoutesWrapper

	method   string
	relPath  string
	handlers []gin.HandlerFunc
}

func (w *GinWrapper) Routes() *GinRoutesWrapper {
	return &GinRoutesWrapper{wrapper: w, routes: w.engine}
}

func (w *GinRoutesWrapper) Group(relativePath string) *GinRoutesWrapper {
	return &GinRoutesWrapper{wrapper: w.wrapper, routes: w.routes.Group(relativePath), defaultHandler: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) Use(middleware ...gin.HandlerFunc) *GinRoutesWrapper {
	defHandler := langext.ArrCopy(w.defaultHandler)
	defHandler = append(defHandler, middleware...)
	return &GinRoutesWrapper{wrapper: w.wrapper, routes: w.routes, defaultHandler: defHandler}
}

func (w *GinRoutesWrapper) GET(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodGet, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) POST(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodPost, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) DELETE(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodDelete, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) PATCH(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodPatch, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) PUT(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodPut, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) OPTIONS(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodOptions, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) HEAD(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: http.MethodHead, relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) COUNT(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: "COUNT", relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRoutesWrapper) Any(relativePath string) *GinRouteBuilder {
	return &GinRouteBuilder{routes: w, method: "*", relPath: relativePath, handlers: langext.ArrCopy(w.defaultHandler)}
}

func (w *GinRouteBuilder) Use(middleware ...gin.HandlerFunc) *GinRouteBuilder {
	w.handlers = append(w.handlers, middleware...)
	return w
}

func (w *GinRouteBuilder) Handle(handler WHandlerFunc) {

	middlewareNames := langext.ArrMap(w.handlers, func(v gin.HandlerFunc) string { return nameOfFunction(v) })
	handlerName := nameOfFunction(handler)

	w.handlers = append(w.handlers, Wrap(w.routes.wrapper, handler))

	methodName := w.method

	if w.method == "*" {
		methodName = "ANY"
		for _, method := range anyMethods {
			w.routes.routes.Handle(method, w.relPath, w.handlers...)
		}
	} else {
		w.routes.routes.Handle(w.method, w.relPath, w.handlers...)
	}

	w.routes.wrapper.routeSpecs = append(w.routes.wrapper.routeSpecs, ginRouteSpec{
		Method:      methodName,
		URL:         w.relPath,
		Middlewares: middlewareNames,
		Handler:     handlerName,
	})
}

func (w *GinWrapper) NoRoute(handler WHandlerFunc) {
	w.engine.NoRoute(Wrap(w, handler))

	w.routeSpecs = append(w.routeSpecs, ginRouteSpec{
		Method:      "ANY",
		URL:         "[NO_ROUTE]",
		Middlewares: nil,
		Handler:     nameOfFunction(handler),
	})
}

func nameOfFunction(f any) string {

	n := runtime.FuncForPC(reflect.ValueOf(f).Pointer()).Name()

	split := strings.Split(n, "/")
	if len(split) == 0 {
		return ""
	}

	return split[len(split)-1]
}
